import React from 'react';
import * as THREE from 'three';
import { EventDispatcher } from 'three';
import { GLTFLoader } from 'three/examples/jsm/loaders/GLTFLoader';
import { DRACOLoader } from 'three/examples/jsm/loaders/DRACOLoader';
import { TransformControls } from 'three/examples/jsm/controls/TransformControls';
import { OrbitControls } from 'three/examples/jsm/controls/OrbitControls';

export class SceneHandler extends EventDispatcher {

    constructor() {
        super();
        this.raycaster = new THREE.Raycaster();
        this.scene = undefined;
        this.camera = undefined;
        this.renderer = undefined;
        this.finalVectorArray = [];
        this.mixerList = [];
        this.animationsList = [];
        this.modelsList = new THREE.Group();
        this.gltfLoader = new GLTFLoader().setDRACOLoader(new DRACOLoader().setDecoderPath('../js/draco/gltf/'));
        this.textureLoader = new THREE.TextureLoader();
        this.clock = new THREE.Clock();
        this.transforControl = undefined;
        this.currentBlindModel = undefined;
        this.zoomvalue = 0.1;
        this.defaultZoom = undefined;
        this.orbitControl = undefined;
        this.listeners = [];
        this.configObj = {
            "foldIn": false,
            "foldOut": false,
            "zoomIn": true,
            "zoomOut": true,
            "open": true,
            "close": true,
        };
        this.imageSize = { x: 400, y: 300 };
        this.imageHalfWidth = this.imageSize.x / 2;
        this.imageHalfHeight = this.imageSize.y / 2;
        this.imageResolution = 75;

        this.orbitMinDistance = 0.8;
        this.orbitMaxDistance = 5;
    }


    removeModelList = () => {
        this.scene.remove(this.modelsList);
    }

    setupScene = (three) => {
        this.scene = three.scene;
        this.camera = three.camera;
        this.renderer = three.gl;
        this.renderer.shadowMap.enabled = true; // Enable shadows
        this.renderer.shadowMap.type = THREE.PCFSoftShadowMap;
        this.raycaster = three.raycaster;
        this.scene.add(this.modelsList);
        this.defaultZoom = this.camera.position.clone().z;
        this.orbitControl = new OrbitControls(this.camera, this.renderer.domElement);
        this.orbitControl.enableRotate = false;
        this.orbitControl.minDistance = this.orbitMinDistance;
        this.orbitControl.maxDistance = this.orbitMaxDistance;
        this.orbitControl.update();
        this.animate();

    }

    updateCameraView = () => {
        const defaultMinDistanceFactor = 0.8;
        const defaultMaxDistanceFactor = 5;
        const imageSizeX = this.imageSize.x || 400; 
        const widthScalingFactor = imageSizeX / 400;
        const minDistanceFactor = defaultMinDistanceFactor * widthScalingFactor;
        const maxDistanceFactor = defaultMaxDistanceFactor * widthScalingFactor;
    
        this.orbitMinDistance = minDistanceFactor;
        this.orbitMaxDistance = maxDistanceFactor;
        this.orbitControl.minDistance = this.orbitMinDistance;
        this.orbitControl.maxDistance = this.orbitMaxDistance;
        this.camera.position.z = 3 * widthScalingFactor; 
        this.defaultZoom =  this.camera.position.z;
        this.camera.updateProjectionMatrix();
        this.orbitControl.update();
    }

    addEventListener = (type, callback) => {
        this.listeners.push({ type, callback });
    }

    dispatchEvent = (event) => {
        this.listeners.forEach(listener => {
            if (listener.type === event.type) {
                listener.callback(event);
            }
        });
    }

    setBaseModel = () => {

    }

    addLight = () => {
        const directionalLight = new THREE.DirectionalLight(0xffffff, 0.2);
        directionalLight.position.set(0, 3, -2);
        directionalLight.castShadow = true;
        directionalLight.shadow.camera.near = 0.1;
        directionalLight.shadow.camera.far = 30;
        directionalLight.target.position.set(0, 0, 0);
        this.scene.add(directionalLight);
    }

    addShadowPlane = () => {
        const planeGeometry = new THREE.PlaneGeometry(5, 2);
        const loader = new THREE.TextureLoader();

        // this.planeMaterial = new THREE.ShadowMaterial({ opacity: 0.2 });
        this.planeMaterial = new THREE.MeshPhongMaterial({ color: 0x00ff00 });

        this.planeMaterial.side = THREE.DoubleSide;
        const plane = new THREE.Mesh(planeGeometry, this.planeMaterial);
        plane.rotateX(-1.07);
        plane.position.set(0, -1.2913098317295773, 0.6473203737076866);
        plane.scale.y = 0.5;
        plane.scale.x = 0.8;
        plane.receiveShadow = true;
        this.scene.add(plane);
    }


    fetchData = async (image) => {
        try {
            const blob = await fetch(image).then((response) => response.blob());
            const reader = new FileReader();

            reader.onload = (event) => {
                const base64Image = event.target.result.split(',')[1];

                const apiUrl = 'https://mukesh3444-ai-detect.hf.space/?logs=container';
                const payload = {
                    img: base64Image,
                };
                const jsonPayload = JSON.stringify(payload);
                const headers = {
                    'Content-Type': 'application/json',
                };
                const requestOptions = {
                    method: 'POST',
                    headers: headers,
                    body: jsonPayload,
                };

                fetch(apiUrl, requestOptions, {
                    mode: 'no-cors',
                })
                    .then((response) => {
                        if (!response.ok) {
                            throw new Error(`HTTP Error! Status: ${response.status}`);
                        }
                        return response.json();
                    })
                    .then((data) => {
                        this.createBlindsData(data.windows_vertices);
                    })
                    .catch((error) => {
                        console.error('Error:', error);
                    });
            };

            reader.readAsDataURL(blob);
        } catch (error) {
            console.error('Error fetching or processing API data:', error);
        }
    };

    sortXYPointsForRectangle = (points) => {
        let centerX = 0;
        let centerY = 0;
        for (const [x, y] of points) {
            centerX += parseInt(x);
            centerY += parseInt(y);
        }
        centerX /= points.length;
        centerY /= points.length;

        points.sort(([x1, y1], [x2, y2]) => {
            const angle1 = Math.atan2(y1 - centerY, x1 - centerX);
            const angle2 = Math.atan2(y2 - centerY, x2 - centerX);
            return angle1 - angle2;
        });
    }

    createBlindsData = (apiData) => {
        this.finalVectorArray = [];
        let scale = 0.02;
        let arrayPixelPoints = JSON.parse(JSON.stringify(apiData));
        const tempArray = [];
        for (let i = 0; i < arrayPixelPoints.length; i++) {
            let arrayvector = [];
            this.sortXYPointsForRectangle(arrayPixelPoints[i]);
            arrayPixelPoints[i].push(arrayPixelPoints[i][0]);

            arrayPixelPoints[i].forEach((item) => {
                let vector = new THREE.Vector3(
                    (parseInt(item[0]) - this.imageHalfWidth) / this.imageResolution,
                    -(parseInt(item[1]) - this.imageHalfHeight) / this.imageResolution,
                    0.05
                );
                arrayvector.push(vector);
            });

            tempArray.push([arrayvector[0], arrayvector[2], scale]);
            // console.log(arrayvector[0], arrayvector[2], scale);
        }

        this.finalVectorArray = tempArray;
        this.updateCameraView();
        this.createBlinds();

    };

    scaleYAxis = (status) => {
        this.modelsList.children.forEach(child => {
            if (status) child.scale.y += 0.1;
            else child.scale.y -= 0.1;
        })
    }

    scaleXAxis = (status) => {
        this.modelsList.children.forEach(child => {
            if (status) child.scale.x += 0.1;
            else child.scale.x -= 0.1;
        })
    }

    scaleNegativeYAxis = (status) => {
        this.modelsList.children.forEach(child => {
            if (status) {
                child.scale.y += 0.1; 
                child.position.y -= 0.19; 
            } else {
                child.scale.y -= 0.1; 
                child.position.y += 0.19; 
            }
        });
    };

    positionXAxis = (status) => {
        this.modelsList.children.forEach(child => {
            if (status) child.position.x += 0.1;
            else child.position.x -= 0.1;
        })
    }

    playAnimation = (type, status) => {
        this.animationsList.forEach(item => {
            let aniName = item.name.toLowerCase();
            if (status && !aniName.includes("unfold")) {
                if (!item.paused && aniName.includes(type)) {
                    item.timeScale = 1;
                    item.play();
                    setTimeout(() => {
                        item.paused = true;
                    }, 1800);
                }

                else if (item.paused && aniName.includes(type)) {
                    item.timeScale = 1;
                    item.paused = false;
                    setTimeout(() => {
                        item.paused = true;
                    }, 500);
                }


            }

            else if (!status && !aniName.includes("unfold")) {
                item.timeScale = -1;
                if (!item.paused && aniName.includes(type)) {
                    item.play();
                    setTimeout(() => {

                        item.paused = true;
                    }, 1800);
                }

                else if (item.paused && aniName.includes(type)) {
                    item.paused = false;
                    setTimeout(() => {
                        item.paused = true;
                    }, 500);
                }


            }
        })
    }

    changeTexture = (image) => {

        const newTexture = this.textureLoader.load(image);
        this.modelsList.traverse(item => {
            if (item.isMesh) {
                if (item.name.includes("Fabric") || item.name.includes("Wood") || item.name.includes("wood")) {
                    item.material.color = null;
                    item.material.map = newTexture;
                    item.material.needsUpdate = true;
                }
            }
        });
    }

    animate = () => {
        requestAnimationFrame(this.animate);
        if (this.mixerList.length > 0) {

            let delta = this.clock.getDelta();
            this.mixerList.forEach((mixer) => {
                mixer.update(delta);
            });
        }

    }

    changeSampleRoom = (image) => {
        this.fetchData(image);
        const backgroundTexture = this.textureLoader.load(image);
        let bgmesh = this.scene.getObjectByName("bg");
        if (!bgmesh) return;
        bgmesh.material.map = backgroundTexture;
        bgmesh.material.needsUpdate = true;
    }

    updateBlinds = () => {
        if (this.modelsList.children.length > 0) {

            this.modelsList.children.map((item, index) => {
                let pointObj = this.finalVectorArray[index];
                let point = pointObj[0];
                let lastPt = pointObj[1];
                let scale = pointObj[2];

                const boundingBox = new THREE.Box3().setFromObject(item);
                const size = new THREE.Vector3();
                boundingBox.getSize(size);
                let actualWidth = lastPt.x - point.x;
                let actualHeight = lastPt.y - point.y;

                item.scale.set(actualWidth / size.x, Math.abs(actualHeight / size.y), scale);
                item.position.set((point.x + ((lastPt.x - point.x) / 2)) + 0.05, lastPt.y, point.z + 0.08);
                if (index == 0) item.position.x = item.position.x - 0.1;
            });
        }
    }

    resetZoom = () => {
        this.camera.position.z = this.defaultZoom;
        this.orbitControl.update();
    }

    zoom = (isZoomIn) => {
        if (isZoomIn) {
            if (this.camera.position.z > this.orbitMinDistance) {
                this.camera.position.z -= this.zoomvalue;
            }
        }
        else {
            if (this.camera.position.z <= this.defaultZoom + 2) {
                this.camera.position.z += this.zoomvalue;
            }
        }

    }

    emptyGroup = () => {
        while (this.modelsList.children.length > 0) {
            let child = this.modelsList.children[0];
            this.modelsList.remove(child);
        }
    }

    deleteAllBlindDetails = () => {
        this.emptyGroup();
        this.animationsList = [];
        this.mixerList = [];
    }

    changeBlind = (glbModel) => {
        this.currentBlindModel = glbModel;
        this.createBlinds();
    }

    loadModel = (item, index) => {
        return new Promise((resolve, reject) => {
            let windowScene = undefined;
            let mixer = undefined;
            let animations = undefined;

            this.gltfLoader.load(this.currentBlindModel, (gltf) => {
                let point = item[0];
                let lastPt = item[1];
                let scale = item[2];

                windowScene = gltf.scene;

                window.scene = this.scene;
                windowScene.traverse(item => {
                    if (item.isMesh) {
                        item.castShadow = true;
                        item.receiveShadow = true;
                    }
                })
                windowScene.castShadow = true;
                windowScene.receiveShadow = true;
                animations = gltf.animations;
                if (animations.length > 0) {
                    this.configObj["foldIn"] = false;
                    this.configObj["foldOut"] = false;
                    animations.map(anime => {
                        if (anime.name.includes("Fold") || anime.name.includes("fold")) {
                            this.configObj["foldIn"] = true;
                            this.configObj["foldOut"] = true;
                        }

                    })
                }
                mixer = new THREE.AnimationMixer(gltf.scene);
                this.mixerList.push(mixer);
                const boundingBox = new THREE.Box3().setFromObject(windowScene);
                const size = new THREE.Vector3();
                boundingBox.getSize(size);
                let actualWidth = lastPt.x - point.x;
                let actualHeight = lastPt.y - point.y;
                windowScene.scale.set(actualWidth / size.x, Math.abs(actualHeight / size.y), scale);
                windowScene.position.set((point.x + ((lastPt.x - point.x) / 2)) + 0.05, lastPt.y, point.z + 0.08);
                if (index == 0) windowScene.position.x = windowScene.position.x - 0.1;
                if (index == 1) windowScene.position.x = windowScene.position.x + 0.08;
                if (animations.length > 0) {
                    animations.forEach(clip => {
                        let action = mixer.clipAction(clip);
                        action.name = clip.name;
                        action.loop = THREE.LoopOnce;
                        action.clampWhenFinished = true;
                        this.animationsList.push(action);
                    })

                }
                this.modelsList.add(windowScene);
                resolve();
            });
        });
    }

    addModels = (vectoryArray) => {
        this.finalVectorArray = vectoryArray;
        this.createBlinds();
    }


    createBlinds = () => {
        this.deleteAllBlindDetails();
        Promise.all(this.finalVectorArray.map((item, index) => this.loadModel(item, index)))
            .then(() => {
                this.dispatchEvent({ type: "loadedData", data: true });
                this.dispatchEvent({ type: "uiOptions", data: this.configObj });
            })
            .catch(error => {
                console.error('Error loading models:', error);
            });
    }

    saveAsImage = (productName) => {
        this.renderer.render(this.scene, this.camera);
        const strMime = "image/jpeg";
        const imgData = this.renderer.domElement.toDataURL(strMime);
        if (productName === 'pdfImageUrl') {
            return imgData
        }
        const link = document.createElement('a');
        link.href = imgData;
        link.download = `${productName}.jpg`;
        link.click();
    }

}